//------------------------------------------------------------------------------
//
//  npgldraw.c
//
// ANTZ - 3D data viewer with realtime real-world interaction based on NPE.
//
// ANTZ is hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
// Written in 2010-2011 by Shane Saxon - contact@saxondigital.net
//
// Additional code contributed by: Mark Buchanan and Lucas Erickson.
//
// To the extent possible under law, the author(s) have dedicated all copyright 
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// Released under the CC0 license, which is GPL compatible.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software (license file named COPYING.txt). If not, see
// http://creativecommons.org/publicdomain/zero/1.0/
//
//------------------------------------------------------------------------------

#include "npgldraw.h"

#include "npgeometry.h"

#include "../../npdata.h"
#include "../../npctrl.h"
#include "../npglut.h"



//MB//
extern int selectedID;
int pickPass;				//zz move it to global data ptr, debug
int selectedTree;
int selectTreeID;

//MB-Transp
//Array of Sorted Pins (pointers)
#define MAXPINS 10000
NPnodePtr Pins[MAXPINS];
int numPins;

void DrawDefault (NPnodePtr node);
void DrawCamera (NPnodePtr node);
void DrawSurface (NPnodePtr node);
void DrawPoints (NPnodePtr node);
void DrawPin (int selected, NPnodePtr node, void* dataRef);
void DrawPinChild (NPnodePtr node, void* dataRef);
void DrawVideo (NPnodePtr node);
void DrawGrid (NPnodePtr node);




//------------------------------------------------------------------------------
void npInitGLDraw(void* dataRef)
{
//	InitTorus (dataRef);
	pickPass = false;			//zz global init
	numPins = 0;				//MB-Transp
}


//------------------------------------------------------------------------------
void DrawDefault (NPnodePtr node)
{
	return;
}


// draw an icon at the location of the camera in the scene
//------------------------------------------------------------------------------
void DrawCamera (NPnodePtr node)
{
	// draw icon of camera in the scene, but not the current viewport camera
	return;
}


//------------------------------------------------------------------------------
void DrawSurface (NPnodePtr node)
{
	// draw ground
	glPushMatrix();

	// position the node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);
	glRotatef (node->rotate.angle, node->rotate.x, node->rotate.y, node->rotate.z);
	glScalef (node->scale.x, node->scale.y, node->scale.z);

	glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);

	glBegin(GL_QUADS);
		glVertex3f (-100.0f, -100.0f, 0.0f);
		glVertex3f (-100.0f,  100.0f, 0.0f);
		glVertex3f ( 100.0f,  100.0f, 0.0f);
		glVertex3f ( 100.0f, -100.0f, 0.0f);
	glEnd();

	glPopMatrix();
}


//------------------------------------------------------------------------------
void DrawPoints (NPnodePtr node)
{
	return;
}


//MB//

// converts the ID value to an RGB value for picking objects by color
//------------------------------------------------------------------------------
void IDtoRGB( int id, int* r, int* g, int* b )
{
	int remain = id;
	
	*r = remain / (256*256);	remain = remain - *r * (256*256);
	*g = remain / 256;			remain = remain - *g * 256;
	*b = remain;
}

int RGBtoID( int r, int g, int b )
{
	int id;
	id = (256*256)*r + 256*g + b;

	return id;
}


void PickPass( int picking )
{
	pData data = (pData) npGetDataRef();

	pickPass = picking;
//	data->io.picking = picking;			/debug, zz
}

int npGetPickPass()
{
	return pickPass;
}

//MB-End//

//------------------------------------------------------------------------------
void DrawPin (int currentNode, NPnodePtr node, void* dataRef)
{
	int i = 0;
	
	int branchLevel = 0;

	pData data = (pData) dataRef;

	NPnodePtr nodeChild = data->map.activeNode;

	//MB//	
	int idRed, idGrn, idBlu;
	int remain = 0;
	idRed = idGrn = idBlu = 0;

	//SelectById. Select from selectedID
	if (pickPass)
	{
		IDtoRGB( node->id, &idRed, &idGrn, &idBlu );	//Convert nodeID to RGB triplet

		if( selectedID == node->id )
		{
			//select the proper root 
			selectedTree = true;
			currentNode = true;
			data->map.nodeRootIndex = selectTreeID;
			data->map.selectedIndex = data->map.nodeRootIndex;
		}
	}
	//MB-End//
	

	branchLevel = nodeChild->branchLevel;

	glPushMatrix();

	// position the node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);
	glRotatef (node->rotate.angle, node->rotate.x, node->rotate.y, node->rotate.z);
	glScalef (node->scale.x, node->scale.y, node->scale.z);

		//MB//
	//Main Color for Rendering Pin
	if( pickPass )
	{
//		glDisable( GL_TEXTURE_2D ); //MB-TEXTURE
		glEnable(GL_DEPTH_TEST);	//MB-Transp debug zz
		glDisable(GL_BLEND);		//MB-Transp
		glDisable(GL_LIGHTING);
		glColor4ub( idRed, idGrn, idBlu, 255 );	//ID of Node is encoded on the Pixel! In RGB Triplet   (Range: 0-  16.7 Million)
	}
	else
	{
		//MB-TEXTURE
//		glEnable( GL_TEXTURE_2D );
//		glEnable(GL_DEPTH_TEST);	//MB-Transp
		if (data->io.gl.alphaMode)
			glEnable(GL_BLEND);		//MB-Transp
		else
			glDisable(GL_BLEND);
		glEnable(GL_LIGHTING);
		glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);
	}
	//-

//	glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);  //debug, zz
	glLineWidth(node->lineWidth);															//zz

//		npGLTexture (node->geometry, node->textureID);

	if (node->geometry == kNPprimitivePin)
	{
		npGLSurface (node->geometry, node->textureID);	
		glTranslatef (0.0f, 0.0f, 5.0f);			//should this be here, debug, zz
	}
	else
	{
		glTranslatef (0.0f, 0.0f, 5.0f);
		npGLSurface (node->geometry, node->textureID);
	}
//		glDisable ( GL_TEXTURE_2D );

	glLineWidth(1.0f);	//zz

	if (node->selected || currentNode)
	{
		glPushMatrix();

		if (currentNode)
		{
			// slide down a little for scaling up to surround the pin
			glTranslatef (0.0f, 0.0f, -5.10f);
			glScalef (1.15f, 1.15f, 1.15f);


			if ( node->selected )
			{
				glColor4f ( 0.8f, 0.8f, 0.0f, 1.0f);	// draw selected yellow
				npGLSurface (kNPprimitiveWirePin, 0);
			}
			else if (branchLevel == 0)
			{
				glColor4f ( 1.0f, 0.0f, 0.0f, 1.0f);	// draw node red
				npGLSurface (kNPprimitiveWirePin, 0);
			}
		//	else
		//		glTranslatef (0.0f, 0.0f, 5.0f);		// move up to draw other

			if (branchLevel > 0)
			{
				if (nodeChild->selected)
					glColor4f ( 0.8f, 0.8f, 0.0f, 1.0f);	// draw selected yellow
				else
					glColor4f ( 1.0f, 0.0f, 0.0f, 1.0f);	// draw selection red
			}
			glTranslatef (0.0f, 0.0f, 5.0f);
			glScalef (1.0f, 1.0f, 0.66f);
			npGLSurface(kNPprimitiveWirePin, 0);		// 2nd one will be above
		}
		else if (node->selected)
		{
			glTranslatef (0.0f, 0.0f, -5.10f);
			glScalef (1.05f, 1.05f, 1.05f);

			glColor4f ( 0.8f, 0.8f, 0.0f, 1.0f);		// draw selected as yellow

			npGLSurface (kNPprimitiveWirePin, 0);
		}

		glPopMatrix();
	}

	//should call DrawNode, that calls draw pin, move this to drawnodes...   debug, zz
	for (i=0; i < node->childCount; i++)
		DrawPinChild (node->child[i], dataRef);

	glPopMatrix();
}

// first draw child here, then break it up, draw nodes should iterate down, debug, zz
// the tree branches and take care of selected glows, etc...
// this routine should be generic regardless of branch level
//------------------------------------------------------------------------------
void DrawPinChild (NPnodePtr node, void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	NPpinPtr nodeData = (NPpinPtr) node->data;


	//MB//
	int idRed = 0, idGrn = 0, idBlu = 0;


	// SelectById. Select from selectedID
	if (pickPass && selectedID == node->id )
	{
		selectedTree = true;
		data->map.nodeRootIndex = selectTreeID;
		data->map.selectedIndex = data->map.nodeRootIndex;				//zz end
	}
	//MB-End//


	if (node->branchLevel > 1)
	{
		glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
		glRotatef (node->rotate.angle, 0.0f, 1.0f, 0.0f);		//x	debug zz

		// translate * scale keeps object 
		glTranslatef (1.5f + node->translate.x * node->scale.x, 0.0f, 0.0f);

		glRotatef (node->rotateRad.y * 57.29578f, 0.0f, 0.0f, 1.0f);		//y, debug zz
		glRotatef (node->rotateRad.z * 57.29578f, 1.0f, 0.0f, 0.0f);		//y, debug zz
		glScalef ( 0.5f * node->scale.x, 0.5f * node->scale.y, 0.5f * node->scale.z);
	}
	else
	{
		glRotatef (node->rotateRad.y * 57.29578f, 0.0f, 0.0f, 1.0f);		//y, debug zz
		glScalef ( node->scale.x, node->scale.y, node->scale.z);	//root child
	}

	//MB//
	//Main Color for Rendering Pin
	if( pickPass )
	{
		glDisable(GL_BLEND);		//MB-Transp
		glDisable(GL_LIGHTING);

		//Convert nodeID to RGB triplet
		IDtoRGB( node->id, &idRed, &idGrn, &idBlu );

		glColor4ub( idRed, idGrn, idBlu, 255 );	//ID of Node is encoded on the Pixel! In RGB Triplet   (Range: 0-  16.7 Million)
	}
	else
	{
		if (data->io.gl.alphaMode)
			glEnable (GL_BLEND);		//MB-Transp
		else
			glDisable (GL_BLEND);

		glEnable (GL_LIGHTING);
		glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);
	}

	// draw the node unless hidden
	if (!node->hide)
	{
		glLineWidth(node->lineWidth);
		npGLSurface (node->geometry, node->textureID);
		glLineWidth(1.0f);
	}

	// draw a selection status wireframes slightly scaled up on top...
	if (node->selected)
	{	
		glPushMatrix();
			// draw nodeID color on pickPass to allow picking on the wireframe
			if( !pickPass )
				glColor4f (0.8f, 0.8f, 0.0f, 1.0f);			//draw yellow
			glScalef (1.03f, 1.03f, 1.03f);
			npGLSurface (kNPglutWireTorus, 0);
		glPopMatrix();
	}

	// draw a wireframe slightly scaled up on top...
	if (node == data->map.selectedNode)
	{	
		glPushMatrix();
			// draw nodeID color on pickPass to allow picking on the wireframe
			if( !pickPass )
				glColor4f (1.0f, 0.0f, 0.0f, 1.0f);			//draw red
			glScalef (1.04f, 1.04f, 1.04f);
			npGLSurface (kNPglutWireTorus, 0);
		glPopMatrix();
	}

	//draw points slightly scaled up on top...
	if (node->pointSize > 0.0f)
	{
		//draw points slightly scaled up on top...		//add feature, debug zz
	}

	
	// draw the child nodes
	for (i=0; i < node->childCount; i++)
	{
		glPushMatrix();
			DrawPinChild (node->child[i], data);
		glPopMatrix();	
	}
}


//------------------------------------------------------------------------------
void DrawVideo (NPnodePtr node)
{
	return;
}


//------------------------------------------------------------------------------
void DrawGrid (NPnodePtr node)
{
	int			i = 0;
	int			j = 0;
	int			k = 0;
	GLfloat		x = 0.0f;
	GLfloat		y = 0.0f;
	GLfloat		z = 0.0f;
	NPfloatXYZ	length;

	NPgridPtr grid = (NPgridPtr) node->data;

	length.x = grid->length.x * grid->segments.x;
	length.y = grid->length.y * grid->segments.y;
	length.z = grid->length.z * grid->segments.z;

	//MB-TEXTURE
//	glDisable( GL_TEXTURE_2D );

	glPushMatrix();

	// position the node
	glTranslatef (node->translate.x, node->translate.y, node->translate.z);
	glRotatef (node->rotate.angle, node->rotate.x, node->rotate.y, node->rotate.z);
	glScalef (node->scale.x, node->scale.y, node->scale.z);

	glTranslatef (length.x / -2.0f, length.y / -2.0f, 0.0f);

	glColor4ub (node->color.r, node->color.g, node->color.b, node->color.a);

	glDisable (GL_LIGHTING);

	glBegin (GL_LINES);

		glLineWidth (node->lineWidth);

		for (i=0; i <= grid->segments.z; i++)
		{
			for (j=0; j <= grid->segments.x; j++)
			{
				glVertex3f (x, 0.0f, z);
				glVertex3f (x, length.y, z);
				glVertex3f (0.0f, y, z);
				glVertex3f (length.x, y, z);
				x += grid->length.x;
				y += grid->length.y;
			}
			x = 0.0f;
			y = 0.0f;
			z += grid->length.z;
		}

		glLineWidth (1.0f);

	glEnd();

	glEnable (GL_LIGHTING);

	glPopMatrix();
}


// draw 2D overlays including grids, text, messages, highlight objects, etc...
//------------------------------------------------------------------------------
void npDrawHUD (void* dataRef)
{
	return;
}


// sets the node RGB color based on index, preserves existing alpha value
//------------------------------------------------------------------------------
void SetIndexColor (NPubyteRGBA *color, int *colorIndex)
{	
	int index = *colorIndex;
	
	GLubyte colorPallet[kNPcolorIndexSize][3] = {	
											{255, 255, 255},
											{0,   255, 0},
											{255,   0, 0},
											{0,     0, 255},
											{255, 255, 0},
											{152,   0, 255},
											{255, 168, 0},
											{0,   255, 255},
											{255,   0, 255},
											{0,   153, 0},
											{173, 255, 202},
											{255, 180, 255},
											{0,   152, 255},
											{185, 255, 0},
											{152,   0, 0},
											{50,  101, 101},
											{127, 127, 255},
											{185, 153, 102},
											{197, 82,  0},
											{127, 127, 127}  };
											
	//make sure colorIndex is inbounds, if not, then put it inbounds																			
	if (index >= kNPcolorIndexSize)
		while (index >= kNPcolorIndexSize)
			index -= kNPcolorIndexSize;		//subtracts pallete sized increments
	
	if (index < 0)
		while (index < 0)
			index += kNPcolorIndexSize;		//adds pallete sized increments

	//now set the color
	color->r = colorPallet[index][0];
	color->g = colorPallet[index][1];
	color->b = colorPallet[index][2];

	*colorIndex = index;
}




//MB-Transp



//------------------------------------------------------------------------------
//Distance between two points
float distBetween( NPfloatXYZ vecA, NPfloatXYZ vecB )
{
	float x = vecA.x - vecB.x;
	float y = vecA.y - vecB.y;
	float z = vecA.z - vecB.z;
	
	float len = x*x + y*y + z*z;
	len = (float)sqrt( len );
	return len;
}

//------------------------------------------------------------------------------
//QSort Compare function
int comparePinsDistanceFromCamera (const void* a, const void* b)
{
	NPnodePtr* pinAptr = (NPnodePtr*)a;
	NPnodePtr* pinBptr = (NPnodePtr*)b;
	NPnodePtr pinA = *pinAptr;
	NPnodePtr pinB = *pinBptr;
	if( pinA == NULL && pinB == NULL )	//Null element in pinA & pinB: Nothing or Grid or Camera
		return 0;						//These are "equal". Both should be at the beginning of the list.
	if( pinA == NULL )					//Null element in pinA: Nothing or Grid or Camera
		return -1;						//pinA is at beginning of list... therefore pinA is    "less than" pinB. Return -1
	if( pinB == NULL )					//Null element in pinB: Nothing or Grid or Camera
		return +1;						//pinB is at beginning of list... therefore pinA is "greater than" pinB. Return +1
	
	if( pinA->distFromCamera   >  pinB->distFromCamera )	//pinA is furter from camera than pinB.
		return -1;											//Therefore should be at the beginning of the list. ie "less than"... Return -1
	if( pinA->distFromCamera   <  pinB->distFromCamera )	//pinA is closer from camera than pinB.
		return +1;											//Therefore should be at the end of the list.    ie "greater than"... Return +1
	if( pinA->distFromCamera  ==  pinB->distFromCamera )	//pinA & pinB are exactly the same distance from the camera. (impossible-ish)
		return 0;											//Return 0;

	//return ( *(int*)a - *(int*)b );
	return -1;
}

//------------------------------------------------------------------------------



void npDrawNodes (void* dataRef)
{
	int i = 0;
	int selected = 0;
	float distFromCamera = 0.0f;

	pData data = (pData) dataRef;
	NPnodePtr node = NULL;
	
	NPfloatXYZ pinPos;
	NPfloatXYZ camPos;
	NPnodePtr camNode = data->map.node[1];		//Camera

	int nodeRootCount = data->map.nodeRootCount;


	camPos = camNode->translate;

	//i[0]: Nothing
	//i[1]: Camera
	//i[2]: Grid
	//i[3]: First Pin
	for( i = 0; i < nodeRootCount; i++ )
	{
		if( i < 3 )	//Nothing, Camera, Grid
		{
			Pins[i] = NULL;
			continue;
		}
		
		node = data->map.node[i];
		pinPos = node->translate;
		distFromCamera = distBetween( camPos, pinPos );
		node->distFromCamera = distFromCamera;

		if (data->map.nodeRootIndex == i)
			node->nodeRootIndexSelected = true;
		else
			node->nodeRootIndexSelected = false;
		
		Pins[i] = node;
	}
	//QuickSort based on Distance From Camera
	qsort( Pins, nodeRootCount, sizeof(NPnodePtr), comparePinsDistanceFromCamera );	
	

	for( i=0; i < nodeRootCount; i++ )
	{
		if( i < 3 )
			node = data->map.node[i];
		else
			node = Pins[i];
		
		//used for mouse picking
		selectTreeID = i;
		node = data->map.node[selectTreeID];
		if (data->map.nodeRootIndex == selectTreeID)
			selectedTree = true;
		else
			selectedTree = false;

		if( node->nodeRootIndexSelected == true )
			selected = true;
		else
			selected = false;
		
		if (!node->hide)			// don't draw hidden nodes
			switch (node->type)
		{
			case kNodeDefault	: DrawDefault(node); break;
			case kNodeCamera	: DrawCamera (node); break;
			case kNodeSurface	: DrawSurface(node); break;
			case kNodePoints	: DrawPoints (node); break;
			case kNodePin		: DrawPin (selected, node, data); break;
			case kNodeVideo		: DrawVideo  (node); break;
			case kNodeGrid		: DrawGrid   (node); break;
		}
	}
}

//MB-LABEL
void DrawSimpleTorus(void)
{
	npGLSurface (kNPglutSolidTorus, 0);
}

